import numpy as np

from boss.bo.acq.base import BaseAcquisition


class Explore(BaseAcquisition):
    """Purely exploring acquisition function.

    Doesn't take any parameters.
    """
    def evaluate(self, x):
        """
        Compute the exploration acqfn at location x.

        Parameters
        ----------
        x : ndarray
          Input to the acquisition function. 2D array containing data with
          'float' type.

        Returns
        -------
        f_acq : ndarray
          Acquisition function evaluated at 'x'. 2D array containing data with
          'float' type.
        """
        mean, var = self.model.predict(x)
        f_acqu = -np.sqrt(var)
        return f_acqu

    def evaluate_with_gradient(self, x):
        """
        Compute the exploration acqfn and its gradient at location x.
        Gradients are required to apply L-BFGS-B minimizer.

        Parameters
        ----------
        x : ndarray
          Input to the acquisition function. 2D array containing data with
          'float' type.

        Returns
        -------
        f_acq, df_acq : tuple of ndarray
          Acquisition function and it's gradient, evaluated at 'x'. Tuple of
          2D arrays, containing data with 'float' type.
        """
        mean, sd, dmean, dsd = self.model.predict_mean_sd_grads(x)
        f_acqu = -sd
        df_acqu = -dsd
        return f_acqu, df_acqu.T

    @property
    def has_gradient(self):
        """Property to inform whether acquisition function has a gradient.

        Returns
        -------
        bool
        """
        return True



import numpy as np
from scipy.optimize import minimize
from scipy.stats import gamma


def gammaparams(q1, q2, p1=0.1, p2=0.9):
    """A function for parametrizing gamma distributions by specifying two
    quantiles.

    Relevant math can be found in "Determining distribution parameters from
    quantiles" by John D. Cook, 2010.

    :param q1: p1-quantile
    :param q2: p2-quantile
    :param p1: lower percentage (0 < p1 < 1)
    :param p2: higher percentage (0 < p2 < 1)
    """
    # Require the arguments to be positive and q1 < q2, p1 < p2 <1
    if q1 <= 0 or q2 <= q1:
        return (None, None)
    if p1 <= 0 or p2 <= p1 or p2 >= 1:
        return (None, None)

    def f(a):
        return np.abs(
            gamma.ppf(p2, a, scale=1) / gamma.ppf(p1, a, scale=1) - q2 / q1
        )

    shape = minimize(f, 1, bounds=[[0.1, None]], method="L-BFGS-B").x
    rate = gamma.ppf(p1, shape, scale=1) / q1

    return (shape[0], rate[0])



import os
import copy
from collections import UserDict

import os
import copy
import numpy as np

import boss.keywords as bkw
import boss.io.parse as parse
import boss.io.dump as dump
from boss.bo.acq.ei import EI
from boss.bo.acq.elcb import ELCB
from boss.bo.acq.exploit import Exploit
from boss.bo.acq.explore import Explore
from boss.bo.acq.lcb import LCB
from boss.bo.acq.cost import AdditiveCost, DivisiveCost
from boss.utils.distributions import gammaparams
from boss.bo.userfunc import UserFunc


class Settings(UserDict):
    """Reads, interprets and defines the code internal settings based on input.
    """

    def __init__(self, keywords, f=None):
        super().__init__()

        # Non-keyword attributes.
        self.is_rst = False

        self.dir = os.getcwd()
        self.acqfn = None
        self.costfn = None
        self.user_keys = keywords.keys()

        # Before any other keywords are set, we assign values to all keywords
        # with independent defaults.
        self.set_independent_defaults(only_missing=False)

        # Update with BOSS keywords passed to init
        self.update(keywords)
        if keywords.get("bounds") is None:
            raise ValueError(
                "Keyword 'bounds' has to be defined by the user"
            )

        # correct keyword types
        self.correct()

        # Handle the user function: if a function is passed directly we take that,
        # otherwise fall back to the function specified by the userfn keyword.
        self.f = None
        if f is None:
            if self["userfn"] is not None:
                f = bkw.func_from_keyword(self["userfn"])
                self.f = UserFunc(f, self.dim)
        else:
            self["userfn"] = bkw.func_to_keyword(f)
            self.f = UserFunc(f, self.dim)

        # Handle cost function: take either directly passed function or
        # function specified by costfn keyword
        if isinstance(self["costfn"], str):
            self.costfn = bkw.func_from_keyword(self["costfn"])
        else:
            if callable(self["costfn"]):
                self.costfn = self["costfn"]
                self["costfn"] = bkw.func_to_keyword(self.costfn)

        # Set default values for dependent keywords if they
        # have not yet been specified.
        self.set_acqfn(self["acqfn_name"])
        self.set_dependent_defaults(only_missing=True)

        # Set RNG seed if specified.
        # TODO: Propagate this seed to GPy to eliminate all randomness.
        if self["seed"] is not None:
            np.random.seed(self["seed"])

    @classmethod
    def from_file(cls, file_path):
        """Factory method for Constructing a Settings object from a boss input file.

        Parameters
        -----------
        file_path: str, Path
            Path to the input file.

        Returns
        -------
        Settings
            Settings object generated using the input file.
        """
        input_data = parse.parse_input_file(file_path, skip="results")
        settings = cls(input_data['keywords'])
        settings.is_rst = input_data['is_rst']
        return settings

    def copy(self):
        return copy.deepcopy(self)

    def set_independent_defaults(self, only_missing=True):
        """Sets default values for independent keywords. """
        if not only_missing:
            for cat in bkw.get_copied_categories():
                self.update(cat)
        else:
            for cat in bkw.get_copied_categories():
                for key, val in cat.items():
                    if self.get(key) is None:
                        self[key] = val


    def set_dependent_defaults(self, only_missing=True):
        """Sets default values for keywords that depend on other keywords. """
        should_update = lambda key: self.get(key) is None or only_missing is False

        if should_update("periods"):
            self["periods"] = self["bounds"][:, 1] - self["bounds"][:, 0]
        if should_update("iterpts"):
            self["iterpts"] = int(15 * self.dim ** 1.5)
        if should_update("min_dist_acqs"):
            self["min_dist_acqs"] = 0.01 * min(self["periods"])

        # Default initial hyperparameters.
        if should_update("thetainit"):
            if self["ynorm"]:
                diff = 1.0 # when observations are range-normalized
            else:
                diff = self["yrange"][1] - self["yrange"][0]
            self["thetainit"] = [0.5 * diff]  # sig
            for i in range(self.dim):  # lengthscales
                if self["kernel"][i] == "stdp":  # pbc
                    self["thetainit"].append(np.pi / 10)
                else:  # nonpbc
                    self["thetainit"].append(self["periods"][i] / 20)

        # Default hyperparameter constraints.
        if self["thetaprior"] is not None:
            if should_update("thetabounds"):
                self["thetabounds"] = [
                    [self["thetainit"][0] / 1000.0, self["thetainit"][0] * 1000.0]
                ]  # variance
                for i in range(self.dim):  # lengthscale
                    self["thetabounds"].append(
                        [
                            self["thetainit"][i + 1] / 100.0,
                            self["thetainit"][i + 1] * 100.0,
                        ]
                    )
                self["thetabounds"] = np.array(self["thetabounds"])

            # Default hyperparameter priors.
            if should_update("thetapriorpar"):
                if self["ynorm"]:
                    diff = 1.0 # when observations are range normalised
                else:
                    diff = self["yrange"][1] - self["yrange"][0]    
                if self["thetaprior"] == "gamma":

                    # Ulpu's heuristic prior
                    shape = 2.00
                    rate = 2.0/(diff/2.0)**2

                    # Original solution, to be tested further
                    #shape, rate = Distributions.gammaparams(
                    #    (diff/4)**2, (10*diff/4)**2, 0.5, 0.99)
                    #shape = 1.0    # NORMALIZATION
                    #rate = 1.5     # NORMALIZATION
                    self["thetapriorpar"] = [[shape, rate]]

                    for i in range(self.dim):
                        if self["kernel"][i] == "stdp":  # pbc
                            shape = 3.3678
                            rate = 9.0204
                        else:  # nonpbc
                            shape, rate = gammaparams(
                                self["periods"][i] / 20, self["periods"][i] / 2
                            )
                        self["thetapriorpar"].append([shape, rate])
                    self["thetapriorpar"] = np.asarray(self["thetapriorpar"])
                else:
                    raise TypeError(
                        f"Unknown options set for thetaprior: {self['thetaprior']}."
                    )

        # Model slice and number of points.
        if should_update('pp_model_slice'):
            if self.dim == 1:
                self["pp_model_slice"] = np.array([1, 1, 50])
            else:
                self["pp_model_slice"] = np.array([1, 2, 25])

    @property
    def dim(self):
        """The dimensionality of the user-supplied objective.

        The number of dimensions is a read-only propery that is
        derived from the bounds provided by the user.

        Returns
        -------
        int
            The dimensionality of the objective.

        """
        return len(self["bounds"])

    @dim.setter
    def dim(self, val):
        raise AttributeError("Cannot modify read-only attribute: dim")

    def set_acqfn(self, name):
        """Initializes the acquisition function based on settings.

        Wraps the acquisition function by a cost function, if
        a cost function is set by the user.
        """
        self["acqfnpars"] = np.atleast_1d(self["acqfnpars"])
        if name != 'explore':
            self.expfn = Explore()
        if name == 'ei':
            if len(self["acqfnpars"]) < 1:
                self["acqfnpars"] = np.array([0.0])
            self.acqfn = EI(self["acqfnpars"])
        elif name == 'elcb':
            self.acqfn = ELCB()
        elif name == 'exploit':
            self.acqfn = Exploit()
        elif name == 'explore':
            self.acqfn = Explore()
            self.expfn = self.acqfn
        elif name == 'lcb':
            if len(self["acqfnpars"]) < 1:
                self["acqfnpars"] = np.array([2.0])
            self.acqfn = LCB(self["acqfnpars"])
        else:
            raise TypeError(f"Unknown acquisition function selected: {name}")

        if callable(self.costfn):
            if self["costtype"] == "add":
                self.acqfn = AdditiveCost(self.acqfn, self.costfn)
                self.expfn = AdditiveCost(self.expfn, self.costfn)
            elif self["costtype"] == "divide":
                self.acqfn = DivisiveCost(self.acqfn, self.costfn)
                self.expfn = DivisiveCost(self.expfn, self.costfn)
            else:
                raise ValueError(f'Unknown costtype {self["costtype"]}.')

    def dump(self, file_path, only_user_keywords=True):
        """Writes the current settings to a boss input file.

        Parameters
        ----------
        fname : Union[str, path]
            Path to the destination file.
        """
        if only_user_keywords:
            keywords = {k: v for k, v in self.items() if k in self.user_keys}
        else:
            keywords = self
        dump.dump_input_file(file_path, keywords)

    def correct(self):
        """Corrects the type and value of certain keywords.

        The user is afforded some laziness defining certain keywords,
        e.g., by providing lists instead of np.arrays.
        """
        # Make sure we're not using deprecated keywords
        self.deprecation_notice()

        # Make sure int and float arrays are np and not python sequences.
        for key, val in self.items():
            cat = bkw.find_category(key)
            cat_type, cat_dim = cat[0], cat[1]
            if cat_dim > 0 and cat_type in [int, float] and val is not None:
                self[key] = np.asarray(val, dtype=cat_type)

        self['bounds'] = np.atleast_2d(self['bounds'])

        kernel = self["kernel"]
        if isinstance(kernel, str):
            self["kernel"] = [kernel]

        if len(self["kernel"]) == 1:
            self["kernel"] *= self.dim

        # Exception to the above rule: glmin_tol is a mixed list [float, int]
        # TODO: break glmin_tol into two keywords in a future release
        glmin_tol = self["glmin_tol"]
        if glmin_tol is not None:
            self["glmin_tol"] = [float(glmin_tol[0]), int(glmin_tol[1])]

    def deprecation_notice(self):
        deprecated = {
            'pp_truef_at_xhats': 'pp_truef_at_glmins',
            'gm_tol': 'glmin_tol',
            'verbosity': None
        }
        msg = ''
        for key, val in deprecated.items():
            if key in self:
                if val is None:
                    msg += f'Keyword {key} has been deprecated and should be removed\n'
                else:
                    msg += f'Keyword {key} has been renamed, use {val} instead\n'

        if len(msg) > 0:
            raise RuntimeError(msg)



import os
import shutil
import numpy as np

from boss.mep.mep import MEP
from boss.mep.space import Space
import boss.io.dump as dump
import boss.io.ioutils as ioutils
import boss.io.parse as parse
from boss.bo.rstmanager import split_rst_data
from boss.pp.pp_main import recreate_bo
from boss.pp.plot import plot_model


class MEPMain:

    def __init__(self, settings, ipt_rstfile, minimafile, main_output):
        # create needed directories
        if os.path.isdir("mep"):
            print("warning: overwriting directory 'mep'")
        shutil.rmtree("mep", ignore_errors=True)
        os.makedirs("mep", exist_ok=True)

        # recreate model and read local minima
        self.get_model(settings, ipt_rstfile, main_output)
        self.get_minima(minimafile, main_output)
        self.get_space(settings)

        # initialize and run
        mep = MEP(
            self.bo,
            self.space,
            self.minima,
            settings["mep_precision"],
            settings["mep_rrtsteps"],
            settings["mep_nebsteps"],
            settings["mep_maxe"],
        )
        mep.run_mep(main_output)

        # write to file
        for path in mep.fullpaths:
            dump.dump_mep(path)

        # plot
        if self.minima.shape[1] == 2:
            self.plot2D(settings, mep)

    def get_model(self, settings, rstfile, main_output):
        input_data = parse.parse_input_file(rstfile)
        rst_data = input_data['rst_data']
        acqs, mod_par = split_rst_data(rst_data, settings.dim)
        self.bo = recreate_bo(
            settings,
            acqs[:, 1:],
            mod_par[mod_par.shape[0] - 1, 1:],
        )

    def get_minima(self, minimafile, main_output):
        self.minima = parse.parse_minima(minimafile)
        self.minima = self.minima[:, :-2]

    def get_space(self, settings):
        bounds = np.transpose(settings["bounds"])
        pbc = np.array(settings["kernel"]) == "stdp"
        if not np.all(settings["periods"] == (bounds[1, :] - bounds[0, :])):
            print("warning: MEP currently assumes periods to match " + "boundlength")
        self.space = Space(bounds, pbc)

    def plot2D(self, settings, mep):
        it = np.max(settings["pp_iters"])
        npts = settings["initpts"] + it
        fname = "postprocessing/data_models/" + "it%.4i_npts%.4i.dat" % (it, npts)
        if not files_ok([fname]):
            print(
                "Model data of the last iteration is required for "
                + "automatic 2D plotting, check\nthe 'pp_models' "
                + "and 'pp_iters' options, then try rerunning postprocessing."
            )
            return
        mdata = ioutils.read_cols(fname, skiprows=2)
        xhat = None
        xnext = None
        minima = self.minima
        truef = None

        plot_model(
            settings,
            "mep/minpaths.png",
            mdata,
            minima=self.minima,
            incl_uncert=False,
            paths=mep.fullpaths,
        )


def files_ok(filenames):
    """
    Checks that the given files exist and can be opened.
    """
    for fname in filenames:
        try:
            f = open(fname, "r")
            f.close()
        except FileNotFoundError:
            print("Could not find file '" + fname + "'")
            return False
    return True



from copy import deepcopy
from typing import *

import numpy as np


def seq_to_array(seq):
    """Performs a item-type sensitive numpy-conversion for a sequence.

    Creates a numpy array from a sequence. The dimensionlity of the output
    array is determined by the type of the items: 0-dimensional items are 
    arranged in a 1d array and array items are stacked vertically.

    Parameters
    ----------
    seq : Sequence
        The sequence which to convert to numpy.

    Returns
    -------
    np.ndarray
        Numpy array with the same contents as the original sequence.
    """
    if len(seq) == 0:
        return np.array([])
    item = seq[0]
    if np.ndim(item) == 0:
        return np.array(seq)
    else:
        return np.vstack(seq)


def wrap_index(ind: int, lim: int) -> int:
    """Wraps a negative index around a limit.

    Performs numpy-style negative index wrapping around a limit value.
    More precisely, if lim is the length of an array then ind = -1 will
    yield lim - 1 and more generally ind < 0 gives lim - ind
    (modulo the limit).

    This function is used to give numpy-like array index wrapping to
    the SparseList class.

    Parameters
    ----------
    ind : int
        The index which to wrap.
    lim : int
        The limit which to wrap around.

    Returns
    -------
    int:
        The wrapped index.
    """
    return max(ind % lim, ind)


def wrap_index_array(ind, lim: int):
    """Like wrap_index, but operates on an index array. """
    return np.maximum(np.asarray(ind) % lim, ind)


class SparseList:
    """A sparse, 1D list structure that stores arbitrary data types.

    This data structure is similar to a dictionary-of-keys sparse matrix
    but can hold arbitrary data types and is strictly 1D. The main purpose
    of this class is to store BOSS results that are only computed at certain
    iterations. 

    The content values can be accessed either through [], which uses a normal index that
    respects the sparse structure of the array, while the values() method takes an index 
    which ignores gaps to access value directly as though they were arranged in a dense array.
    """
    def __init__(self, data=None, default=None) -> None:
        """Creates a new SparseList.

        Parameters
        ----------
        data: dict
            A dict (mapping indices to values) to use as data.
        default: scalar
            A scalar value that is returned for any gaps in the list. Typically
            this is set to None, np.nan or 0.
        """
        self.data = {}
        self._inds_values = []
        self.ind_max = -1
        self.default = default
        if data:
            for ind, val in data.items():
                self[ind] = val

    @classmethod
    def from_array(cls, arr, default=None) -> "SparseList":
        """Creates a SparseList from a normal numpy/python array.

        Parameters
        ----------
        arr : array_like
            The array to convert.
        default : Optional[Any]
            A default value that is treated like empty space in the array. 

        Returns
        -------
        SparseList
            The resulting SparseList.
        """
        self = cls()
        for ind, val in enumerate(arr):
            if val != default:
                self[ind] = val
        return self

    def __len__(self) -> int:
        """Returns the length of the sparse list. """
        return self.ind_max + 1

    def __eq__(self, other: "SparseList") -> bool:
        """Implements the == (equality) operator, for sparse lists. """
        is_eq = (
            self.data == other.data
            and self._inds_values == other._inds_values
            and self.ind_max == other.ind_max
        )
        return is_eq

    def __bool__(self) -> bool:
        """Allows us to treat non-empty sparse lists as truthy. """
        return bool(len(self))

    def copy(self) -> "SparseList":
        """Deepcopies a sparse list. """
        return deepcopy(self)

    def __delitem__(self, ind):
        """Deletes an item at a given index. """
        del self.data[ind]
        ival = self.to_value_index(ind)
        del self._inds_values[ival]

    def __getitem__(self, ind: int) -> Any:
        """Returns the item at the given normal index. """
        length = len(self)
        if not -(length + 1) < ind < length:
            raise IndexError
        ind = wrap_index(ind, length)
        return self.data.get(ind, self.default)

    def __setitem__(self, ind: int, value: Any) -> None:
        """Sets an item at the given normal index. """
        self.ind_max = max(self.ind_max, ind)
        ind = wrap_index(ind, len(self))
        self.data[ind] = value
        self._inds_values.append(ind)

    def from_value_index(self, ival: int) -> int:
        """Converts a value index to a normal index. """
        if abs(ival) > len(self.data):
            raise IndexError
        ival = wrap_index(ival, len(self.data))
        ind = self._inds_values[ival]
        return ind

    def to_value_index(self, ind: int) -> int:
        """Converts a normal index to a value index. """
        length = len(self)
        if not -(length + 1) < ind < length:
            raise IndexError
        ind = wrap_index(ind, length)
        ival = self._inds_values.index(ind)
        return ival

    def to_array(self, include_gaps=False) -> np.ndarray:
        """Converts the sparse list to a dense numpy array. """
        if include_gaps:
            arr = seq_to_array([self.data.get(i) for i in range(len(self))])
        else:
            arr = seq_to_array(list(self.data.values()))
        return arr

    def append(self, value: Any) -> None:
        """Appends a value to the end of the list. """
        self[self.ind_max + 1] = value

    def value(self, ival: int) -> Any:
        """Access an item using a value index which ignores gaps in the list.
        """
        if abs(ival) > len(self.data):
            raise IndexError
        ival = wrap_index(ival, len(self.data))
        ind = self._inds_values[ival]
        return self.data[ind]

    def values(self):
        """Returns the values stored in the list. """
        return self.data.values()

    def keys(self):
        """Returns the keys, i.e. indices to the values. """
        return self.data.keys()

    def items(self):
        return self.data.items()

    def is_default(self, ind):
        """Checks if a given value is the sparse default value. """
        if ind not in self.data:
            return True
        val = self.data[ind]
        if np.ndim(val) == 0:
            return val in [self.default]
        return False

    def __repr__(self) -> str:
        """Returns a string representation of the list. """
        return f"SparseList({str(self.data)})"



import abc
import contextlib
import os
import shutil
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np


def remove_boss_files():
    path = Path.cwd()
    for file_path in [path / "boss.in", path / "boss.out", path / "boss.rst"]:
        if file_path.exists():
            file_path.unlink()
    shutil.rmtree(path / "postprocessing", ignore_errors=True)


class execute_in(contextlib.ContextDecorator):
    def __init__(self, rundir):
        self.rundir = Path(rundir)
        self.origin = Path.cwd()
        super().__init__()

    def __enter__(self):
        os.chdir(self.rundir)

    def __exit__(self, *exc):
        os.chdir(self.origin)


class TestFunc(abc.ABC):
    """Base class for benchmark optimization problems."""

    # Should be truthy if the test function can accept multiple
    # data points, given as rows in an ndarray, and return a corresponding
    # array of evalutations.
    vectorized = False

    def __init__(
        self,
        std: Optional[float] = None,
        seed: Optional[int] = None,
        ndim: Optional[int] = None,
        bounds: Optional[np.ndarray] = None,
    ) -> None:

        # Noisy evaluations
        self.std = std
        self.seed = seed
        if seed:
            np.random.seed(seed)

        # Default value of the bounds attribute
        if bounds is not None:
            bounds = np.asarray(bounds)

        self._bounds = bounds

        # Default value of the ndim attribute
        if ndim:
            self._ndim = ndim
        elif self.bounds is not None:
            self._ndim = self.bounds.shape[0]

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def tags(self) -> dict:
        """Tags to make the function searchable / categorizable."""
        return {}

    @property
    def minima(self) -> List[Tuple[np.ndarray, float]]:
        """Optional, a list of tuples (xmin, ymin) containing any local or global minima."""
        raise NotImplementedError

    @property
    def bounds(self) -> np.ndarray:
        """The bounds of the problem, give row-wise in an ndarray."""
        return self._bounds

    @bounds.setter
    def bounds(self, new_bounds: np.ndarray) -> None:
        new_bounds = np.asarray(new_bounds)
        if self.ndim is not None:
            if new_bounds.shape[0] != self.ndim:
                raise ValueError(
                    f"Specified bounds {new_bounds}"
                    + f"are not consistent with ndim = {self.ndim}"
                )
        self._bounds = new_bounds

    @property
    @abc.abstractmethod
    def fmin(self) -> float:
        """The global minimum value."""
        raise ValueError

    @property
    @abc.abstractmethod
    def xmins(self) -> np.ndarray:
        """The global minima given as rows in a 2d np.ndarray."""
        pass

    @property
    def xmin(self) -> np.ndarray:
        """The global minimum, if a single one exists"""
        if self.xmins.shape[0] == 1:
            return self.xmins[0, :]
        else:
            raise ValueError

    @property
    def frange(self) -> np.ndarray:
        """Optional, the expected range over which f varies.

        Only used in BOSS to initialize hyperparameters."""
        return None

    @abc.abstractmethod
    def eval(self, x: np.ndarray):
        """Implementation of the test function."""
        pass

    def eval_grad(self, x: np.ndarray):
        """Implementation of the test function."""
        pass

    @property
    def ndim(self) -> int:
        """The dimension of the function domain."""
        return self._ndim

    def __call__(self, x: np.ndarray):
        if self.vectorized:
            x = np.atleast_2d(x)
        else:
            x = np.asarray(x)
        y = self.eval(x)
        if self.std:
            y += np.random.normal(0, self.std, size=y.shape)
        return np.squeeze(y)

    def grad(self, x: np.ndarray):
        if self.vectorized:
            x = np.atleast_2d(x)
        else:
            x = np.asarray(x)
        y = self.eval_grad(x)
        if self.std:
            y += np.random.normal(0, self.std, size=y.shape)
        return np.squeeze(y)


def get_test_func(name: str, **kwargs) -> TestFunc:
    """Convenience instantiation of any test function."""
    test_funcs = {
        "forrester": Forrester,
        "grammacylee": GrammacyLee,
        "ackely1": Ackley1,
        "alpine1": Alpine1,
        "alpine2": Alpine2,
        "adjiman": Adjiman,
        "bartelsconn": BartelsConn,
        "beale": Beale,
        "camelthreehump": CamelThreeHump,
        "camelsixhump": CamelSixHump,
        "eggcrate": EggCrate,
        "exponential": Exponential,
        "goldsteinprice": GoldsteinPrice,
        "himmelblau": Himmelblau,
        "himmelvalley": Himmelvalley,
        "periodic": Periodic,
        "rastigrin": Rastigrin,
        "rosenbrock": Rosenbrock,
        "salomon": Salomon,
        "sphere": Sphere,
        "styblinskitang": StyblinskiTang,
        "wolfe": Wolfe,
    }
    name_lower = name.lower()
    if name_lower in test_funcs:
        f = test_funcs[name_lower](**kwargs)
    else:
        raise ValueError(f"Could not find test function {name}.")
    return f


class Forrester(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "multimodal": True,
        "dimension": 1,
    }

    bounds = np.array([[0.0, 1.0]])
    frange = np.array([-10.0, 10.0])
    xmins = np.array([[0.757249]])
    fmin = -6.02074
    ndim = 1

    def eval(self, x):
        return (6 * x - 2) ** 2 * np.sin(12 * x - 4)


class GrammacyLee(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 1,
    }

    bounds = np.array([[0.5, 2.5]])
    frange = np.array([-10.0, 10.0])
    xmins = np.array([[0.548563]])
    fmin = -0.869011
    ndim = 1

    def eval(self, x):
        return 0.5 * np.sin(10 * np.pi * x) / x + (x - 1) ** 4


class Ackley1(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }

    @property
    def bounds(self):
        return np.tile([-35.0, 35.0], (self.ndim, 1))

    @property
    def xmins(self):
        return np.zeros((1, self.ndim))

    @property
    def fmin(self):
        return 0.0

    def eval(self, x):
        d = self.ndim
        val = (
            -20 * np.exp(-0.02 * np.sqrt(1.0 / d * np.sum(x**2, axis=1)))
            - np.exp(1.0 / d * np.sum(np.cos(2 * np.pi * x), axis=1))
            + np.exp(0)
            + 20
        )
        return val


class Alpine1(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": False,
        "dimension": None,
    }
    fmin = 0.0

    @property
    def bounds(self):
        return np.tile([-10.0, 10.0], (self.ndim, 1))

    @property
    def xmins(self):
        return np.zeros((1, self.ndim))

    def eval(self, x):
        val = np.sum(np.abs(x * np.sin(x) + 0.1 * x), axis=1)
        return val


class Alpine2(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }

    def __init__(self, *args, **kwargs):
        if kwargs.get("ndim", None) is None:
            raise ValueError(
                "ndim (postive integer) must be specified for test functions with arbitrary dimension."
            )
        super().__init__(*args, **kwargs)
        if self._bounds is None:
            self._bounds = np.tile([0.0, 10.0], (self.ndim, 1))

    @property
    def xmins(self):
        return np.ones((1, self.ndim)) * 7.917

    @property
    def fmin(self):
        return -(2.808**self.ndim)

    def eval(self, x):
        val = -np.product(np.sqrt(x) * np.sin(x), axis=1)
        return val


class Adjiman(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-1.0, 2.0], [-1.0, 1.0]])
    xmins = np.array([[2, 0.10578]])
    fmin = -2.02181
    ndim = 2

    def eval(self, x):
        val = np.cos(x[:, 0]) * np.sin(x[:, 1]) - x[:, 0] / (x[:, 1] ** 2 + 1)
        return val


class BartelsConn(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": False,
        "dimension": 2,
    }

    bounds = np.array([[-500.0, 500.0], [-500.0, 500.0]])
    xmins = np.array([[0.0, 0.0]])
    fmin = 1.0
    ndim = 2

    def eval(self, x):
        val = (
            np.abs(x[:, 0] ** 2 + x[:, 1] ** 2 + x[:, 0] * x[:, 1])
            + np.abs(np.sin(x[:, 0]))
            + np.abs(np.cos(x[:, 1]))
        )
        return val


class Beale(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-4.5, 4.5]] * 2)
    xmins = np.array([[3.0, 0.5]])
    fmin = 0.0
    ndim = 2

    def eval(self, x):
        val = (
            (1.5 - x[:, 0] + x[:, 0] * x[:, 1]) ** 2
            + (2.25 - x[:, 0] + x[:, 0] * x[:, 1] ** 2) ** 2
            + (2.625 - x[:, 0] + x[:, 0] * x[:, 1] ** 3) ** 2
        )
        return val


class CamelThreeHump(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "multimodal": True,
        "dimension": 2,
    }

    bounds = np.array([[-5.0, 5.0]] * 2)
    xmins = np.array([[0.0, 0.0]])
    fmin = 0.0
    ndim = 2

    def eval(self, x):
        val = (
            2 * x[:, 0] ** 2
            - 1.05 * x[:, 0] ** 4
            + x[:, 0] ** 6 / 6
            + x[:, 0] * x[:, 1]
            + x[:, 1] ** 2
        )
        return val


class CamelSixHump(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-5.0, 5.0]] * 2)
    xmins = np.array([[-0.0898, 0.7126], [0.0898, -0.7126]])
    fmin = -1.0316
    ndim = 2

    def eval(self, x):
        val = (
            (4 - 2.1 * x[:, 0] ** 2 + x[:, 0] ** 4 / 3) * x[:, 0] ** 2
            + x[:, 0] * x[:, 1]
            + (4 * x[:, 1] ** 2 - 4) * x[:, 1] ** 2
        )
        return val


class EggCrate(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    frange = np.array([0.0, 100.0])
    bounds = np.array([[-5.0, 5.0]] * 2)
    xmins = np.array([[0.0, 0.0]])
    fmin = 0.0
    ndim = 2

    def eval(self, x):
        val = (
            x[:, 0] ** 2
            + x[:, 1] ** 2
            + 25 * (np.sin(x[:, 0]) ** 2 + np.sin(x[:, 1]) ** 2)
        )
        return val


class Exponential(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }
    fmin = 1.0

    @property
    def bounds(self):
        return np.tile([-1.0, 1.0], (self.ndim, 1))

    @property
    def xmins(self):
        return np.zeros((1, self.ndim))

    def eval(self, x):
        val = -np.exp(-0.5 * np.sum(x**2, axis=1))
        return val


class GoldsteinPrice(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-2.0, 2.0]] * 2)
    xmins = np.array([[0.0, -1.0]])
    fmin = 3.0
    ndim = 2

    def eval(self, x):
        x0 = x[:, 0]
        x1 = x[:, 1]
        val = (
            1
            + (x0 + x1 + 1) ** 2
            * (19 - 14 * x0 + 3 * x0**2 - 14 * x1 + 6 * x0 * x1 + 3 * x1**2)
        ) * (
            30
            + (2 * x0 - 3 * x1) ** 2
            * (18 - 32 * x0 + 12 * x0**2 + 48 * x1 - 36 * x0 * x1 + 27 * x1**2)
        )
        return val


class Himmelblau(TestFunc):
    """Himmelblau function, has 4 global mins."""

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-5.0, 5.0], [-5.0, 5.0]])
    xmins = np.array(
        [
            [3.0, 2.0],
            [-2.805118, 3.131312],
            [-3.779310, -3.283186],
            [3.584428, -1.848126],
        ]
    )
    fmin = 0.0
    ndim = 2

    def eval(self, X):
        x = X[:, 0]
        y = X[:, 1]
        z = (x**2 + y - 11) ** 2 + (x + y**2 - 7) ** 2
        return z


class Himmelvalley(TestFunc):
    """Modified Himmelblau function with one global min."""

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-5.0, 5.0], [-5.0, 5.0]])
    xmins = np.array([[-4.00035, -3.55142]])
    fmin = -1.463295972168
    frange = np.array([0.0, 10.0])
    ndim = 2

    def eval(self, X):
        x = X[:, 0]
        y = X[:, 1]
        z = 0.01 * ((x**2 + y - 11) ** 2 + (x + y**2 - 7) ** 2 + 20 * (x + y))
        return z


class Periodic(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }
    frange = np.array([0.0, 5])
    fmin = 0.9

    def __init__(self, *args, **kwargs):
        if kwargs.get("ndim", None) is None:
            raise ValueError(
                "ndim (postive integer) must be specified for test functions with arbitrary dimension."
            )
        super().__init__(*args, **kwargs)
        if self._bounds is None:
            self._bounds = np.tile([-10.0, 10.0], (self.ndim, 1))

    @property
    def xmins(self):
        return np.zeros((1, self.ndim))

    def eval(self, x):
        val = 1 + np.sum(np.sin(x) ** 2, axis=1) - 0.1 * np.exp(-np.sum(x**2, axis=1))
        return val


class Rastigrin(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }
    frange = np.array([0.0, 100])
    fmin = 0.0

    def __init__(self, *args, **kwargs):
        if kwargs.get("ndim", None) is None:
            raise ValueError(
                "ndim (postive integer) must be specified for test functions with arbitrary dimension."
            )
        super().__init__(*args, **kwargs)
        if self._bounds is None:
            self._bounds = np.tile([-5.12, 5.12], (self.ndim, 1))

    @property
    def xmins(self):
        return np.zeros((1, self.ndim))

    def eval(self, x):
        val = 10 * self.ndim + np.sum(x**2 - 10 * np.cos(2 * np.pi * x), axis=1)
        return val


class Rosenbrock(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }
    fmin = 0.0

    @property
    def bounds(self):
        return np.tile([-30.0, 30.0], (self.ndim, 1))

    @property
    def xmins(self):
        return np.ones((1, self.ndim))

    def eval(self, x):
        val = np.sum(100 * (x[:, 1:] - x[:, :-1] ** 2) ** 2 + (x[:, :-1] - 1) ** 2)
        return val


class Salomon(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }

    fmin = 0.0

    def __init__(self, *args, **kwargs):
        if kwargs.get("ndim", None) is None:
            raise ValueError(
                "ndim (postive integer) must be specified for test functions with arbitrary dimension."
            )
        super().__init__(*args, **kwargs)
        if self._bounds is None:
            self._bounds = np.tile([-100.0, 100.0], (self.ndim, 1))

    @property
    def xmins(self):
        return np.zeros((1, self.ndim))

    def eval(self, x):
        val = (
            1
            - np.cos(2 * np.pi * np.sqrt(np.sum(x**2, axis=1)))
            + 0.1 * np.sqrt(np.sum(x**2, axis=1))
        )
        return val


class Sphere(TestFunc):

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }

    fmin = 0.0

    @property
    def bounds(self):
        return np.tile([0.0, 10.0], (self.ndim, 1))

    @property
    def xmins(self):
        return np.zeros((1, self.ndim))

    def eval(self, x):
        val = np.sum(x**2, axis=1)
        return val


class StyblinskiTang(TestFunc):

    vectorized = True

    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }
    frange = np.array([-50.0, 100])

    def __init__(self, *args, **kwargs):
        if kwargs.get("ndim", None) is None:
            raise ValueError(
                "ndim (postive integer) must be specified for test functions with arbitrary dimension."
            )
        super().__init__(*args, **kwargs)
        if self._bounds is None:
            self._bounds = np.tile([-5.0, 5.0], (self.ndim, 1))

    @property
    def fmin(self):
        return -39.16599 * self.ndim

    @property
    def xmins(self):
        return -2.903534 * np.ones((1, self.ndim))

    def eval(self, x):
        val = 0.5 * np.sum(x**4 - 16 * x**2 + 5 * x, axis=1)
        return val

    def eval_grad(self, x):
        return 2 * x**3 - 16 * x + 2.5


class Wolfe(TestFunc):

    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 3,
    }

    bounds = np.array([[0.0, 2.0]] * 3)
    xmins = np.array([[0.0, 0.0, 0.0]])
    fmin = 0.0
    ndim = 3

    def eval(self, x):
        val = 4.0 / 3 * (x[0] ** 2 + x[1] ** 2 - x[0] * x[1]) ** 0.75 + x[2]
        return val



from boss.bo.acq.base import BaseAcquisition


class DivisiveCost(BaseAcquisition):
    """
    Wraps the acquisition function with a user-defined cost function.

    DivisiveCost considers the cost by dividing the evaluations by a
    cost factor.
    """
    def __init__(self, acqfn, costfn):
        super().__init__()
        self.acqfn = acqfn
        self.costfn = costfn

    def evaluate(self, x):
        """Compute the acquisition function divided by cost at location x.

        Parameters
        ----------
        x : ndarray
            Input to the acquisition function. 2D array containing data with
            'float' type.

        Returns
        -------
        ndarray
            Acquisition function divided by cost at location x. 2D array
            containing data with 'float' type.
        """
        cost, _ = self.costfn(x)
        return self.acqfn(x) / cost

    def evaluate_with_gradient(self, x):
        """Compute the acquisition function divided by cost and
        its gradient at location x.

        Parameters
        ----------
        x : ndarray
            Input to the acquisition function. 2D array containing data with
            'float' type.

        Returns
        -------
        tuple of ndarray
            Acquisition function divided by cost and its gradient at
            location x. 2D array containing data with 'float' type.
        """
        f_acqu, df_acqu = self.acqfn.evaluate_with_gradient(x)
        cost, dcost = self.costfn(x)
        acqfn_with_cost = f_acqu / cost
        acqfn_with_cost_grad = (df_acqu*cost - f_acqu*dcost) / cost**2
        return acqfn_with_cost, acqfn_with_cost_grad

    @property
    def model(self):
        """Property, wrapped acquisition function (wrapped by cost)
        should access the same model as the unwrapped acquisition function.

        Returns
        -------
        Model
            BaseModel used for the optimization.
        """
        return self.acqfn.model

    @model.setter
    def model(self, new_model):
        """Setter for the model property.

        Parameters
        ----------
        new_model : Model
            BaseModel used for the optimization.
        """
        self.acqfn.model = new_model


class AdditiveCost(BaseAcquisition):
    """
    Wraps the acquisition function with a user-defined cost function.
    With that, the user can add 'weak' constrains on the search space.

    AdditiveCost considers the cost by adding a cost factor to the evaluations.
    """
    def __init__(self, acqfn, costfn):
        super().__init__()
        self.acqfn = acqfn
        self.costfn = costfn

    def evaluate(self, x):
        """Compute the acquisition function with added cost at location x.

        Parameters
        ----------
        x : ndarray
            Input to the acquisition function. 2D array containing data with
            'float' type.

        Returns
        -------
        ndarray
            Acquisition function with added cost at location x. 2D array
            containing data with 'float' type.
        """
        cost, _ = self.costfn(x)
        return self.acqfn(x) + cost

    def evaluate_with_gradient(self, x):
        """Compute the acquisition function with added cost and
        its gradient at location x.

        Parameters
        ----------
        x : ndarray
            Input to the acquisition function. 2D array containing data with
            'float' type.

        Returns
        -------
        tuple of ndarray
            Acquisition function with added cost and its gradient at
            location x. 2D array containing data with 'float' type.
        """
        f_acqu, df_acqu = self.acqfn.evaluate_with_gradient(x)
        cost, dcost = self.costfn(x)
        acqfn_with_cost = f_acqu + cost
        acqfn_with_cost_grad = df_acqu + dcost
        return acqfn_with_cost, acqfn_with_cost_grad

    @property
    def model(self):
        """Property, wrapped acquisition function (wrapped by cost)
        should access the same model as the unwrapped acquisition function.

        Returns
        -------
        Model
            BaseModel used for the optimization.
        """
        return self.acqfn.model

    @model.setter
    def model(self, new_model):
        """Setter for the model property.

        Parameters
        ----------
        new_model : Model
            BaseModel used for the optimization.
        """
        self.acqfn.model = new_model



"""
Base class for acquisition functions used in Bayesian optimization.

To add a new acquisition function:
  1. Create a class that inherits from BaseAcquisition and implement all
     methods, that are decorated with @abstractmethod.
  2. Modify 'set_acqfn' in settings.py module accordingly.
"""
import numpy as np
from abc import ABC, abstractmethod

from boss.utils.minimization import Minimization


class BaseAcquisition(ABC):
    """
    Base class for acquisition functions used in Bayesian optimization.

    This class is intended to be subclassed. It contains functionalities for
    evaluating and minimizing an acquisition function.
    """
    def __init__(self, model=None):
        self._model = model

    @abstractmethod
    def evaluate(self, x):
        """Abstract method. Evaluates the acquisition function at x and
        returns its value.

        Parameters
        ----------
        x : ndarray
            Location to evaluate acquisition function at

        Returns
        -------
        f_acq : ndarray
          Acquisition function evaluated at 'x'. 2D array containing data with
          'float' type.
        """

    @abstractmethod
    def evaluate_with_gradient(self, x):
        """
        Abstract method. Evaluates the acquisition function at x and
        returns its value and gradient.

        Parameters
        ----------
        x : ndarray
            Location to evaluate acquisition functions at.

        Returns
        -------
        tuple of ndarray
          Acquisition function evaluation and it's gradient at 'x'.
          2D arrays with 'float' type.
        """

    def __call__(self, x):
        return self.evaluate(x)

    @property
    def model(self):
        """Property, acquisition functions require access to the model.

        Returns
        -------
        Model
            BaseModel used for the optimization.
        """
        return self._model

    @model.setter
    def model(self, new_model):
        """Setter for the model property.

        Parameters
        ----------
        new_model : Model
            BaseModel used for the optimization.
        """
        self._model = new_model

    def minimize(self, bounds, optimtype='score'):
        """
        Minimizes the acquisition function to find the next
        sampling location 'x_next'.

        Parameters
        ----------
        dim : int
          Dimension of space of the user function.
        bounds : ndarray
          Bounds of used variables, defining the space of the user function.
          2D array containing data with 'float' type.

        Returns
        -------
        ndarray
          Array containing found minimum with 'float' type.
        """
        if optimtype == 'score':
            gmin = Minimization.minimize_using_score(
                self.evaluate_with_gradient, bounds)
        else:
            # Calculate number of local minimizers to start.
            # 1. Estimate number of local minima in the surrogate model.
            estimated_numpts = self.model.estimate_num_local_minima(bounds)
            # 2. Increase estimate to approximate number of local minima in
            # acquisition function. Here we assume that increase is
            # proportional to estimated number of local minima per dimension.
            dim = len(bounds)
            minima_multiplier = 1.7
            estimated_numpts = (minima_multiplier ** dim) * \
                estimated_numpts
            num_pts = min(len(self.model.X), int(estimated_numpts))
            # Minimize acqfn to obtain sampling location
            gmin = Minimization.minimize_from_random(
                self.evaluate_with_gradient,
                bounds,
                num_pts=num_pts,
                )
        return np.atleast_1d(np.array(gmin[0]))



import numpy as np

from boss.bo.acq.base import BaseAcquisition


class ELCB(BaseAcquisition):
    """GP-Lower Confidence Bound acquisition function with increasing
    exploration.

    Doesn't take any parameters. The exploration weight is given by

    weight = sqrt( 2*log[ ( i^((dim/2) + 2)*pi^(2) ) / ( 3*0.1 ) ] )

    The implementation is based on the following papers

    (1) N. Srinivas, A. Krause, S. M. Kakade, and M. Seeger.  Gaussian process
    optimization in the bandit setting: No regret and experimental design.
    Proc. ICML, 2010

    (2) E. Brochu, V. M. Cora, and N. de Freitas. A tutorial on Bayesian
    optimization of expensive cost functions, with application to active user
    modeling and hierarchical reinforcement learning. arXiv:1012.2599, 2010,

    where the delta parameter introduced in Brochu et al. has been set to 0.1.
    """
    def __init__(self):
        super().__init__()

    def evaluate(self, x):
        """
        Compute the ELCB acqfn at location x.

        Parameters
        ----------
        x : ndarray
          Input to the acquisition function. 2D array containing data with
          'float' type.

        Returns
        -------
        f_acq : ndarray
          Acquisition function evaluated at 'x'. 2D array containing data with
          'float' type.
        """
        weight = self._get_explore_weight()
        mean, var = self.model.predict(x)
        f_acqu = mean - weight * np.sqrt(var)
        return f_acqu

    def evaluate_with_gradient(self, x):
        """
        Compute the ELCB acqfn and its gradient at location x.
        Gradients are required to apply L-BFGS-B minimizer.

        Parameters
        ----------
        x : ndarray
          Input to the acquisition function. 2D array containing data with
          'float' type.

        Returns
        -------
        f_acq, df_acq : tuple of ndarray
          Acquisition function and it's gradient, evaluated at 'x'. Tuple of
          2D arrays, containing data with 'float' type.
        """
        weight = self._get_explore_weight()
        mean, sd, dmean, dsd = self.model.predict_mean_sd_grads(x)
        f_acqu = mean - weight * sd
        df_acqu = dmean - weight * dsd
        return f_acqu, df_acqu.T

    def _get_explore_weight(self):
        """Calculate exploration weight for ELCB acquisition function.

        The exploration weight depends on the amount of samples and the
        dimension. The derivation is described in the Article, cited in
        the class docstring.

        Returns
        -------
        int
            Exploration weight.
        """
        N_data = self.model.X.shape[0]
        dim = self.model.dim
        upstairs = N_data**((dim/2) + 2) * np.pi**2
        downstairs = 3*0.1
        weight = np.sqrt(2*np.log10(upstairs/downstairs))
        return weight

    @property
    def has_gradient(self):
        """Property to inform whether acquisition function has a gradient.

        Returns
        -------
        bool
        """
        return True
